%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Functional Generalized Structured Component Analysis (PMET-957R2)
% Written in MATLAB R2014a
% Last modified: April 15, 2016
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Example code for analyzing gait data (gait.mat) using functional GSCA.
% Running create_gait_data.m will create gait.mat file. 
% You should download and install FDA toolbox
% (http://www.psych.mcgill.ca/misc/fda/downloads/FDAfuns/Matlab/) before
% you run this program.

%% USER SPECIFICATION NEEDED 
% Users should specify the followings: 
% - nbt: number of bootstrap samples
% - cl: confidence level 
% - filepath_data: the directory where your data file is stored. 
% - filepath_result: the directory where your result file is stored. 
% - myfilename: data file name 
% - Z0: cell object for data matrix; each cell represents each set; within
% each cell, each row indicates each individual and each column
% indicates each indicator variable 
% - names: the name of each set (or component) 
% - indnames: the name of each indicator variable in each set 
% - IB: model specification of path coefficients 
% - dtype: type of each set (1: functional, 2:multivariate, 3:dummy)

cl = 0.95;
nbt = 500;

%% load data file
filepath_data = 'D:\RESEARCH\Projects\FGSCA\data\';
myfilename = 'gait.mat';

filepath_result = 'D:\RESEARCH\Projects\FGSCA\results\';
mkdir(filepath_result);

load (strcat(filepath_data,myfilename));
% fdobj_gait_mean: functional data object containing the mean gait trajectories (force under left foot)
% fdobj_gait_std: functional data object containing the standard deviation of gait trajectories (force under left foot)
% demodata:
% - column1: Height(meter)
% - column2: Weight(kg)
% - column3: HY
% - column4: UPDRS
% - column5: UPDRS_motor
% - column6: TUAG (Timed Up and Go test)
% - column7: Speed

%% define the model
% observed variables for each set
ind = [1:83]; % include all 83 patients in the analysis
Z0{1} = demodata(ind,[1,2]); % body size
Z0{2} = demodata(ind,[3,4,5]); % set 1: severity of parkinson's disease
Z0{3} = demodata(ind,[6,7]); % set 2: gait function
Z0{4} = fdobj_gait_mean(ind); % set 4: mean gait trajectory

% name of each set
names  = {'body size','PD','gaitfun','gait mean'};
% each name should be less than 10 characters

% name of each variable in each set 
indnames{1} = {'Height','Weight'};
indnames{2} = {'HY','UPDRS','UPDRS_m'}; % indicators in set1
indnames{3} = {'TUG','speed'}; % indicators in set2
indnames{4} = {}; % functional object
% each name should be less than 10 characters

% path coefficients
IB = [1 0 99 99;
    0 1 99 99;
    0 0 0 99;
    0 0 0 0];
% K by K matrix for the specification of path coefficients, where K
% indicates the number of sets of variables (or functions). (k,j)th element
% of IB indicates the path coefficient from kth set to jth set. The value
% of 99 indicates that the corresponding path coefficient is to be
% estimated. Other values will be treated as user-specified fixed values
% for the corresponding path coefficients. (k,k)th element of IB should be
% set to 1 if kth set is an exogenous set.

% type of each set
dtype = [2 2 2 1];
% 1: functional data object
% 2: multivariate
% 3: dummy variable

%% cross-validation
nf = 5;
plambdaw = 10.^[1:1:5]; % possible values of smoothing parameters for weight
plambdac = 10.^[1:1:5]; % possible values of smoothing parameters for loading

% [flambdaw,flambdac,PE,NE,index_p] = fgsca_cv(dtype,Z0,IB,nf,plambdaw,plambdac);
% Uncomment the above line and comment the following line if you want to perform a cross-validation to
% determine lambda_w and lambda_c.
flambdaw = 10^3; flambdac = 10^3;

%% USER SPECIFICATION NOT NEEDED 
% Do not change the codes given below.

%% main analysis
disp('============================');
disp('main analysis');
disp('============================');
[fB,fc,fw,fGamma,FIT,fobjf] = fgsca_main(dtype,Z0,IB,flambdaw,flambdac,1);

%% Bootstrapping
[UPPERB,LOWERB,UPPERC,LOWERC,UPPERW,LOWERW,C,W,BB] = fgsca_bt(dtype,Z0,IB,flambdaw,flambdac,fw,cl,nbt);

%% Print result
clc; close all;
currenttime = clock;
fname = strcat(filepath_result,'result_fgsca_',num2str(currenttime(1)),'_',num2str(currenttime(2),'%02d'),num2str(currenttime(3),'%02d'),'_',num2str(currenttime(4),'%2d'),num2str(currenttime(5),'%2d'),'.txt');
fid = fopen(fname,'w+');
fprintf(fid,'============================\n');
fprintf(fid,'Functional GSCA results\n');
fprintf(fid,'============================\n\n');

fprintf(fid,'Date: %2d/%2d/%4d %02d:%02d\n',currenttime(2),currenttime(3),currenttime(1),currenttime(4),currenttime(5));
fprintf(fid,'Data file: %s\n\n',strcat(filepath,myfilename));
fprintf(fid,'Smoothing parameters:\n');
fprintf(fid,'- log10(lambda_w) = %4.4f\n',log10(flambdaw));
fprintf(fid,'- log10(lambda_c) = %4.4f\n\n',log10(flambdac));

fprintf(fid,'Fit:\n');
fprintf(fid,'- FIT total = %1.4f\n',FIT(1));
fprintf(fid,'- FIT measurement model = %1.4f\n',FIT(2));
fprintf(fid,'- FIT structural model = %1.4f\n\n',FIT(3));

fprintf(fid,'-----------------------------------------------\n');
fprintf(fid,'Path coefficients with %2d percent CI:\n',100*cl);
fprintf(fid,'-----------------------------------------------\n');

k = 1;
nlv = length(Z0);
for j = 1:nlv
    for i = 1:nlv
        if IB(i,j)==99
            fprintf(fid,'%-10s --> %-10s : %1.4f (%1.4f, %1.4f)\n',names{i},names{j},fB(i,j),LOWERB(k),UPPERB(k)); 
            k = k + 1;
        end
    end
end

nt = 500;
fprintf(fid,'\n-----------------------------------------------\n');
fprintf(fid,'Weights with %2d percent CI:\n',100*cl);
fprintf(fid,'-----------------------------------------------\n');
for i = 1:nlv
    if dtype(i)==1 % functional data object
        basisobj{i} = getbasis(Z0{i});
        range{i} = getbasisrange(basisobj{i});
        time{i} = linspace(range{i}(1),range{i}(2),nt);
        PHI{i} = eval_basis(time{i},basisobj{i},0);
        FW{i} = PHI{i}*fw{i};
        
        figure;
        plot(time{i},FW{i});
        hold on
        plot(time{i},UPPERW{i},':');
        plot(time{i},LOWERW{i},':');
        title(strcat('Weight function : ',names{i}));
        xlabel('Percentage');
        ylabel('Weight');
    else
        fprintf(fid,'\nSET: %s\n',names{i});
        for j = 1:length(fw{i})
            fprintf(fid,'- %-10s : %1.4f (%1.4f, %1.4f)\n',indnames{i}{j},fw{i}(j),LOWERW{i}(j),UPPERW{i}(j));
        end
        
    end
end

fprintf(fid,'\n-----------------------------------------------\n');
fprintf(fid,'Loading with %2d percent CI:\n',100*cl);
fprintf(fid,'-----------------------------------------------\n');

for i = 1:nlv
    if dtype(i)==1 % functional data object
        FC{i} = PHI{i}*fc{i};
        
        figure;
        plot(time{i},FC{i});
        hold on
        plot(time{i},UPPERC{i},':');
        plot(time{i},LOWERC{i},':');
        title(strcat('Loading function : ',names{i}));
        xlabel('Percentage');
        ylabel('Loading');
    else
        fprintf(fid,'\nSET: %s\n',names{i});
        for j = 1:length(fc{i})
            fprintf(fid,'- %-10s : %1.4f (%1.4f, %1.4f)\n',indnames{i}{j},fc{i}(j),LOWERC{i}(j),UPPERC{i}(j));
        end
        
    end
end

fclose(fid);
fprintf('The analysis result has been saved on %s.\n',fname);
open(fname);