function [UPPERB,LOWERB,UPPERC,LOWERC,UPPERW,LOWERW,C,W,Bvec] = fgsca_bt(dtype,Z0,IB,flambdaw,flambdac,fw,cl,nbt)
% Functional GSCA bootstrapping code written in MATLAB R2014a
% last modified: Feb 29, 2016
%
% [UPPERB,LOWERB,UPPERC,LOWERC,UPPERW,LOWERW,C,W,Bvec] = fgsca_bt(dtype,Z0,IB,flambdaw,flambdac,fw,fc,cl,nbt)
% Arguments:
% dtype     ... type of each set of data; if there are K sets of variables/
%               functions, dtype should be a K by 1 vector. The kth element
%               of this vector indicates the type of the kth set:
%               - 1: functional data object
%               - 2: multivariate
%               - 3: dummy variable
% Z0        ... a cell object of data. Z0{k} contains the kth set of variables
%               /functions
% IB        ... K by K matrix for the specification of path coefficients.
%               The (k,j)th element of IB indicates the path coefficient from jth
%               set to kth set. The value of 99 indicates that the corresponding
%               path coefficient is to be estimated. Other values will be treated
%               as user-specified fixed values for the corresponding path coefficients.
% fc        ... the estimated loading vectors (or functions) in main
%               analysis
% flambdaw  ... lambda_w to use
% flambdac  ... lambda_c to use
% cl        ... confidence level (default = 0.95; 95% confidence interval)
% nbt       ... number of bootstrap samples (default = 500)
%
% Returns:
% UPPERB    ... upper limits for path coefficient matrix
% LOWERB    ... lower limits for path coefficient matrix
% UPPERC    ... upper limits for loading vector (or the vectors of
%               coefficients for basis functions for loading functions)
% LOWERC    ... lower limits for loading vectors (or the vector of
%               coefficients for basis functions for loading functions)
% UPPERW    ... upper limits for weight vector (or the vectors of
%               coefficients for basis functions for weight functions)
% LOWERW    ... lower limits for weight vectors (or the vector of
%               coefficients for basis functions for weight functions)

if nargin < 8
    nbt = 500;
end

if dtype(1)==1
    nobs = size(getcoef(Z0{1}),2);
else nobs = size(Z0{1},1);
end
nlv = size(Z0,2);

% nobs: number of subjects (N)
% nlv: number of sets (K)

PHI = {}; BC = {}; BW = {}; C = {}; W = {};

clc;
disp('Bootstrapping starts');

for j = 1:nbt
    disp('-----------------------------------');
    disp(['Bootstrap sample = ',num2str(j)]);
    disp('-----------------------------------');
    
    indexbt = []; Zbt = {};
    indexbt = ceil(rand(nobs,1)*nobs);
    
    for i = 1:nlv
        if dtype(i)==1 % functional data object
            Zbt{i} = Z0{i}(indexbt);
        else Zbt{i} = Z0{i}(indexbt,:);
        end
    end
    
    [bfB,bfc,bfw] = fgsca_main(dtype,Zbt,IB,flambdaw,flambdac,1);

    
    nt = 500; % number of time points for evaluation
    if j == 1
        for i = 1:nlv
            if dtype(i)==1
                range{i} = getbasisrange(getbasis(Z0{i}));
                testtime{i} = linspace(range{i}(1),range{i}(2),nt);
                basisobj{i} = getbasis(Z0{i});
                PHI{i} = eval_basis(testtime{i},basisobj{i},0);
            else testtime{i} = zeros(1,size(bfw{i},1));
                PHI{i} = eye(size(bfw{i},1));
            end
            FW{i} = PHI{i}*fw{i};
        end
    end
    
    % change the sign of estimate when necessary
    for i = 1:nlv
        W{i}(:,j) = PHI{i}*bfw{i};
        switch dtype(i)
            case 1
                if sign(W{i}(:,j))'*sign(FW{i})<0
                    bfc{i} = -bfc{i};
                    bfw{i} = -bfw{i};
                    bfB(i,:) = -bfB(i,:);
                    bfB(:,i) = -bfB(:,i);
                end
            case {2,3}
%                 indc = min(find(abs(fc{i})==max(abs(fc{i}))));
%                 if fc{i}(indc)*bfc{i}(indc)<0
                if sign(W{i}(:,j))'*sign(FW{i})<0    
                    bfc{i} = -bfc{i};
                    bfw{i} = -bfw{i};
                    bfB(i,:) = -bfB(i,:);
                    bfB(:,i) = -bfB(:,i);
                end
        end
        BC{i}(:,j) = bfc{i};
        BW{i}(:,j) = bfw{i};
        C{i}(:,j) = PHI{i}*bfc{i};
        W{i}(:,j) = PHI{i}*bfw{i};
    end
    BB(:,:,j) = bfB;
    Bvec(:,j) = bfB(IB==99);
end

% sort bootstrap results
indexupper = ceil(nbt*(cl+(1-cl)/2));
indexlower = floor(nbt*(1-cl)/2);

for i = 1:nlv
    for j = 1:length(testtime{i})
        tempc = []; tempw = [];
        tempc = sort(C{i}(j,:));
        UPPERC{i}(:,j) = tempc(indexupper);
        LOWERC{i}(:,j) = tempc(indexlower);
        tempw = sort(W{i}(j,:));
        UPPERW{i}(:,j) = tempw(indexupper);
        LOWERW{i}(:,j) = tempw(indexlower);
    end
end

for i = 1:size(Bvec,1)
    tempb = [];
    tempb = sort(Bvec(i,:));
    UPPERB(i) = tempb(indexupper);
    LOWERB(i) = tempb(indexlower);
end


