function [flambdaw,flambdac,PE,NE,index_p] = fgsca_cv(dtype,Z0,IB,nf,plambdaw,plambdac,rstart)
% Functional GSCA cross-validation code written in MATLAB R2014a
% last modified: Feb 28, 2016
%
% [flambdaw,flambdac,PE] = fgsca_cv(dtype,Z0,IB,nf,plambdaw,plambdac,rstart)
%
% Arguments:
% dtype     ... type of each set of data; if there are K sets of variables/
%               functions, dtype should be a K by 1 vector. The kth element 
%               of this vector indicates the type of the kth set:
%               - 1: functional data object
%               - 2: multivariate
%               - 3: dummy variable
% Z0        ... a cell object of data. Z0{k} contains the kth set of variables
%               /functions
% IB        ... K by K matrix for the specification of path coefficients. 
%               The (k,j)th element of IB indicates the path coefficient from jth 
%               set to kth set. The value of 99 indicates that the corresponding 
%               path coefficient is to be estimated. Other values will be treated 
%               as user-specified fixed values for the corresponding path coefficients.
% nf        ... number of folders for cross-validation (default = 5)
% plambdaw  ... plausible values for lambda_w to be tested in cross-validation 
%               (default = 10.^[0:1:5])
% plambdac  ... plausible values for lambda_c to be tested in cross-validation 
%               (default = 10.^[0:1:5])
% rstart    ... number of different starting values (default = 1);
% 
% Returns: 
% flambdaw  ... optimal value of lambda_w obtained by cross-validation
% flambdac  ... optimal value of lambda_c obtained by cross-validation
% PE        ... the matrix of prediction errors (column 1:lambda_w, 
%               column 2:lambda_c, column 3:prediction error)
% NE        ... the matrix of the number of times the algorithm does not
%               monotonically decrease (column 1: lambda_2, column 2:
%               lambda_c, column 3: nerror)
% index_p   ... index for permulation of data

clc;
if dtype(1)==1
    nobs = size(getcoef(Z0{1}),2);
else nobs = size(Z0{1},1);
end
nlv = size(Z0,2);
% nobs: number of subjects (N)
% nlv: number of sets (K)

if nargin < 7
    rstart = 1;
end

if nargin < 5
    plambdaw = 10.^[0:1:5]; 
    plambdac = 10.^[0:1:5]; 
end

if nargin < 4
    nf = 5;
end

PE = []; 
NE = []; 

disp('Cross-validation starts.');
% rindex_p = randperm(nobs);
index_p = [1:1:nobs];

for iw = 1:length(plambdaw)
    lambdaw = plambdaw(iw);
    for ic = 1:length(plambdac)
        lambdac = plambdac(ic);
        npe = 0;
        num_errors = 0;
        for n = 1:nf
            
            % make trainset and testset
            testindex = []; trainindex = []; Z0test = {}; Z0train = {};Gammatest= []; Xtest = []; J = [];
            testindex = index_p(floor((n-1) * nobs/nf)+1 : floor(n*nobs/nf));
            trainindex = setdiff(index_p, testindex);
            for i = 1:nlv
                if dtype(i)==1 % functional data
                    Z0test{i} = Z0{i}(testindex);
                    Z0train{i} = Z0{i}(trainindex);
                else
                    Z0test{i} = Z0{i}(testindex,:);
                    Z0train{i} = Z0{i}(trainindex,:);
                end
            end
            nobs_test = length(testindex);
            disp('------------------------------------------------------------------');
            disp(['log10(lambdaw) = ',num2str(log10(lambdaw),'%02d'),', log10(lambdac) = ',num2str(log10(lambdac),'%02d'),', folder = ',num2str(n,'%d')]);
            disp('------------------------------------------------------------------');

            % run fgsca
            [fB,fc,fw,~,~,~,~,fnerror] = fgsca_main(dtype,Z0train,IB,lambdaw,lambdac,rstart);
            
            % test data preprocessing
            for i = 1:nlv
                switch dtype(i)
                    case 1 % functional data object
                        
                        % centering
                        basisobj{i} = getbasis(Z0test{i});
                        coef = getcoef(Z0test{i});
                        center_coef = coef - mean(coef,2)*ones(1,nobs_test);
                        
                        % create matrices required for estimation
                        Xtest{i} = center_coef';
                        J{i} = eval_penalty(basisobj{i},0);
                         
                        % normalization
                        norm = trace(Xtest{i}*J{i}*Xtest{i}');
                        Xtest{i} = sqrt(nobs_test)*Xtest{i}/sqrt(norm); % matrix norm = nobs
                        
                    case 2 % multivariate
                        DZ0 = []; ssZ0 = []; stdZ0 = [];
                        
                        DZ0 = Z0test{i} - ones(nobs_test,1)*mean(Z0test{i}); % variable-wise centering
                        ssZ0 = diag(DZ0'*DZ0)';
                        stdZ0 = sqrt(ssZ0);
                        stdZ0(stdZ0==0)=1e-12;
                        Z{i} = sqrt(nobs_test)*DZ0./(ones(nobs_test,1)*stdZ0); % variable-wise normalization (norm = nobs)
                        Xtest{i} = Z{i};
                        J{i} = eye(size(Xtest{i},2));
                        
                    case 3 % dummy variable
                        Xtest{i} = Z0test{i};% no centering, no normalization
                        J{i} = eye(size(Xtest{i},2));
                end
            end
            
            % calculate prediction error
            for i = 1:nlv
                Gammatest(:,i) = Xtest{i}*J{i}*fw{i};
            end
            
            perror = 0;
            for i = 1:nlv
                perror = perror + trace(Xtest{i}*J{i}*Xtest{i}') - 2*fc{i}'*J{i}*Xtest{i}'*Gammatest(:,i) + fc{i}'*J{i}*fc{i}*Gammatest(:,i)'*Gammatest(:,i);
            end
            ES = Gammatest - Gammatest*fB;
            perror = perror + trace(ES'*ES);
            
            npe = npe + perror;
            num_errors = num_errors + sum(fnerror);
        
        end
        PE = [PE; [log10(lambdaw) log10(lambdac) npe]];
        NE = [NE; [log10(lambdaw) log10(lambdac) num_errors]];
    end
end

pindex = min(find(PE(:,3)==min(PE(:,3))));
flambdaw = 10^PE(pindex,1); % optimal value of lambda_w
flambdac = 10^PE(pindex,2); % optimal vaelu of lambda_c

disp('------------------------------------------------------------------');
disp('Cross-validation result');
disp('------------------------------------------------------------------');
disp(['optimal log10(lambda_w) = ',num2str(log10(flambdaw))]);
disp(['optimal log10(lambda_c) = ',num2str(log10(flambdac))]);


