function [fB,fc,fw,fGamma,FIT,fobjf,fiter,nerror] = fgsca_main(dtype,Z0,IB,lambdaw,lambdac,rstart)
% Main functional GSCA algorithm written in MATLAB R2014a
% last modified: Feb 28, 2016
%
% [fB,fc,fw,fGamma,FIT,fobjf,fiter] = fgsca_main(dtype,Z0,IB,lambdaw,lambdac,rstart)
%  
% Arguments:
% dtype     ... type of each set of data; if there are K sets of variables/
%               functions, dtype should be a K by 1 vector. The kth element 
%               of this vector indicates the type of the kth set:
%               - 1: functional data object
%               - 2: multivariate
%               - 3: dummy variable
% Z0        ... a cell object of data. Z0{k} contains the kth set of variables
%               /functions
% IB        ... K by K matrix for the specification of path coefficients. 
%               The (k,j)th element of IB indicates the path coefficient from jth 
%               set to kth set. The value of 99 indicates that the corresponding 
%               path coefficient is to be estimated. Other values will be treated 
%               as user-specified fixed values for the corresponding path coefficients.
% lambdaw   ... the value of lambda_w to be used for estimation
% lambdac   ... the value of lambda_c to be used for estimation
% rstart    ... number of different starting values (default = 20)
% nerror    ... number of times the algorithm does not monotonically
%               decrease
%
% Returns: 
% fB        ... estimated path coefficients
% fc        ... estimated loadings (or estimated basis function coefficients 
%               for loading functions)
% fw        ... estimated weights (or estimated basis function coefficients 
%               for weight functions)
% fGamma    ... component scores
% FIT       ... fit measures [total fit, measurement model fit, structural model fit]
% fobjf     ... the obtained objective function value at convergence
% fiter     ... the number of iterations

if nargin < 6
    rstart = 20;
end

if dtype(1)==1
    nobs = size(getcoef(Z0{1}),2);
else nobs = size(Z0{1},1);
end
nlv = size(Z0,2);
% nobs: number of subjects (N)
% nlv: number of sets (K)

opts = optimoptions(@fmincon,'Display','Off','TolFun',1e-12,'TolCon',1e-10, 'TolX',1e-12,'Algorithm','interior-point','MaxIter',9999,'MaxFunEvals',99999);
% optimization option

%% raw data preprocessing
for i = 1:nlv
    switch dtype(i)
        case 1 % functional data object
            
            % centering
            basisobj{i} = getbasis(Z0{i});
            coef = getcoef(Z0{i});
            center_coef = coef - mean(coef,2)*ones(1,nobs);
                        
            % create matrices required for estimation
            X{i} = center_coef';
            J{i} = eval_penalty(basisobj{i},0);
            R{i} = eval_penalty(basisobj{i},2);
            JXXJ{i} = J{i}'*X{i}'*X{i}*J{i};
            
            % normalization
            norm = trace(X{i}*J{i}*X{i}');
            X{i} = sqrt(nobs)*X{i}/sqrt(norm); % matrix norm = nobs
            JXXJ{i} = J{i}'*X{i}'*X{i}*J{i};
            nbasis(i) = getnbasis(getbasis(Z0{i}));
            
            Z{i} = fd(X{i}',basisobj{i}); % centered and normalized data
            
        case 2 % multivariate
            DZ0 = []; ssZ0 = []; stdZ0 = [];
            DZ0 = Z0{i} - ones(nobs,1)*mean(Z0{i}); % variable-wise centering as in GSCA
            ssZ0 = diag(DZ0'*DZ0)';
            stdZ0 = sqrt(ssZ0);
            Z{i} = sqrt(nobs)*DZ0./(ones(nobs,1)*stdZ0); % variable-wise centered and normalized data (norm = nobs)
            X{i} = Z{i};
            J{i} = eye(size(X{i},2));
            R{i} = eye(size(X{i},2));
            JXXJ{i} = J{i}'*X{i}'*X{i}*J{i};
            nbasis(i) = size(Z{i},2);
            
        case 3 % dummy variable
            Z{i} = Z0{i}; % no centering, no normalization
            X{i} = Z{i};
            J{i} = eye(size(X{i},2));
            R{i} = eye(size(X{i},2));
            JXXJ{i} = J{i}'*X{i}'*X{i}*J{i};
            nbasis(i) = size(Z{i},2);
    end
end

for k = 1:rstart % rstart: number of random starting values
    
    nerror(k) = 0;
 
    % initialize B
    vecB = IB(:);
    indvecB = find(vecB==99);
    vecB(indvecB,:) = randn(length(indvecB),1);
    B = reshape(vecB,size(IB));
    c = {}; w = {};
    
    % initialize C and W
    for i = 1:nlv
        c{i} = randn(nbasis(i),1);
        w{i} = randn(nbasis(i),1);
        Gamma(:,i) = X{i}*J{i}*w{i};
        stdi = sqrt(Gamma(:,i)'*Gamma(:,i));
        if dtype(i)==3
            w{i} = 1;
        else w{i} = sqrt(nobs)*w{i}/(stdi); % norm = nobs
        end
        Gamma(:,i) = X{i}*J{i}*w{i};
    end
    
    eps = 1e-6; % convergence criterion
    maxit = 500; % maximum number of iterations
    it = 0;
    diff = realmax;
    fobj = realmax;
    obj_old = realmax;
    Gamma_old = Gamma;
    
    E = eye(nlv);
    objf = [];
    
    while (diff > eps)

        it = it + 1; % iteration number
        disp(['iteration = ',num2str(it)]);
        
        % update W (simultaneouly)
        x = []; nlength = [];
        for i = 1:nlv
            x = [x; w{i}];
            nlength(i) = length(w{i});
        end

        exitflag = 100;
        while exitflag == 100
            [xw,fval,exitflag] = fmincon(@(x) myfun(x,nlv,nlength,dtype,X,J,JXXJ,R,B,c,lambdaw),x,[],[],[],[],[],[],@(x) mycon(x,nlv,nlength,JXXJ,nobs,dtype),opts);
            % exitflag
            if exitflag == 0
                warning('fmincon: number of iterations exceeds the maximum. The algorithm will restart with different starting values.')
            elseif exitflag < 0
                warning('fmincon: not converged. The algorithm will restart with different starting values.')
            end
            
            if exitflag <=0
                % restart the algorithm with another starting value
                it = 1; exitflag = 100;
                fobj = realmax;
                obj = realmax;
                diff = realmax;
                obj_old = realmax;
                objf = [];
                
                % initialize B
                vecB = IB(:);
                indvecB = find(vecB==99);
                vecB(indvecB,:) = ones(length(indvecB),1);
                B = reshape(vecB,size(IB));
                c = {}; w = {};
                
                % initialize C and W
                for i = 1:nlv
                    c{i} = rand(nbasis(i),1);
                    w{i} = rand(nbasis(i),1);
                    Gamma(:,i) = X{i}*J{i}*w{i};
                    stdi = std(Gamma(:,i));
                    if dtype(i)==3
                        w{i} = 1;
                    else w{i} = sqrt(nobs)*w{i}/(stdi); % norm = nobs
                    end
                    Gamma(:,i) = X{i}*J{i}*w{i};
                    
                end
            end
        end
        
        for i = 1:nlv
            if i == 1 
                w{i} = xw(1:nlength(1),1);
            else w{i} = xw(sum(nlength(1:i-1))+1:sum(nlength(1:i)),1);
            end
            
            if dtype(i)==3 % dummy
                w{i} = 1;
            end
            
            Gamma(:,i) = X{i}*J{i}*w{i};
        end
        
        for i = 1:nlv
            % update C
            N1 = []; u1 = [];
            if dtype(i)==1
                N1 = (w{i}'*JXXJ{i}*w{i})*J{i} + lambdac*R{i};
            else N1 = (w{i}'*JXXJ{i}*w{i})*J{i};
            end
            u1 = JXXJ{i}*w{i};
            c{i} = inv(N1)*u1;
        end
        
        % update B
        O1 = kron(E,Gamma);
        O2 = O1(:,indvecB);
        vecB(indvecB,1) = O2\Gamma(:);
        B = reshape(vecB,size(IB));
        
        % calculate objective function
        me = 0; se = 0; pe = 0;
        for i = 1:nlv
            me = me + trace(X{i}*J{i}*X{i}') - 2*w{i}'*JXXJ{i}*c{i} + c{i}'*J{i}*c{i}*w{i}'*JXXJ{i}*w{i};
            if dtype(i)==1
                pe = pe + lambdaw*w{i}'*R{i}*w{i} + lambdac*c{i}'*R{i}*c{i};
            else pe = pe;
            end
        end
        ES = Gamma - Gamma*B;
        se = trace(ES*ES');
        obj = me + se + pe;
        objf(it) = obj;
        
        diff = obj_old - obj; % objective function difference
%         disp(['iteration = ',num2str(it,'%03d'),'      objective function decrease = ',num2str(diff)]);
        
        if diff<0 || it>=maxit
            nerror(k) = nerror(k) + 1;
            if diff<0
                warning('Objective function increased. The algorithm will restart with different starting values.')
            end
            
            if it>=maxit
                warning('Maximum number of iterations reached. The algorithm will restart with different starting values.')
            end
            
            % restart the algorithm with another starting value
            it = 0;
            fobj = realmax;
            obj = realmax;
            diff = realmax;
            obj_old = realmax;
            objf = [];

            % initialize B
            vecB = IB(:);
            indvecB = find(vecB==99);
            vecB(indvecB,:) = ones(length(indvecB),1);
            B = reshape(vecB,size(IB));
            c = {}; w = {};
            
            % initialize C and W
            for i = 1:nlv
                c{i} = rand(nbasis(i),1);
                w{i} = rand(nbasis(i),1);
                Gamma(:,i) = X{i}*J{i}*w{i};
                stdi = std(Gamma(:,i));
                if dtype(i)==3
                    w{i} = 1;
                else w{i} = sqrt(nobs)*w{i}/(stdi); % norm = nobs
                end
                Gamma(:,i) = X{i}*J{i}*w{i};
                
            end
        end
        Gamma_old = Gamma;
        obj_old = obj;
    end
    
    if obj < fobj
        fobj = obj;
        fB = B;
        fc = c;
        fw = w;
        fGamma = Gamma;
        fresid = me + se;
        fobjf = []; fobjf = objf;
        fiter = it;
        fse = se;
        fme = me;
    end
    
    % reverse sign if necessary
    for i = 1:nlv
        if sum((sign(fc{i})==-1)) > sum((sign(fc{i})==1))
            fc{i} = -fc{i}; fw{i} = -fw{i};
            fB(i,:) = -fB(i,:); fB(:,i) = -fB(:,i);
        end
    end
    
end % end of k

% FIT
SSme = 0; SSse = 0;
for i = 1:nlv
    if dtype(i)==1% functional
        SSme = SSme + nobs;
        if sum(IB(:,i)==99)~=0
            SSse = SSse + nobs;
        end
    else temp = []; temp = X{i} - ones(nobs,1)*mean(X{i});% multivariate or dummy
        if size(Z0{i})>1
            SSme = SSme + trace(temp'*temp);
        end
        
        temp = []; temp = fGamma(:,i) - ones(nobs,1)*mean(fGamma(:,i));
        if sum(IB(:,i)==99)~=0
            SSse = SSse + trace(temp'*temp);
        end
    end
end

FITtotal = 1 - (fme+fse)/(SSme + SSse);
FITme = 1 - fme/SSme;
FITse = 1 - fse/SSse;
FIT = [FITtotal, FITme, FITse];


